#########################################################
#
#  Application: iterativeStability
#    Version: 1.0 (11 Sept 2018)
#    Author: Roman Jaksik
#            Silesian University of Technology
#            Gliwice, Poland
#    Contact: roman.jaksik@polsl.pl
#
#  Description:
#    This function extends the NormFinder algorithm by 
#    allowing to test various combinations of multiple 
#    features in order to select the most stable set. 
#    In the first iteration the algorithm selects the 
#    most stable feature, in the second step this feature 
#    is averaged with all remaining features, which 
#    are than tested using NormFinder. The best set is 
#    moved to the next step. The algorithm is repeated 
#    k-times, selecting a set of maximum k-features which 
#    combined show the highest stability.
#
#  Required libraries:
#    plyr, ggplot2, openxlsx
#
#  Additional requirements:
#    This software uses the NormFinder function created 
#    in the Department of Molecular Medicine, Aarhus 
#    University Hospital, Denmark
#    The function can be downloaded from:
#    https://moma.dk/files/r.NormOldStab5.txt
#   
#  Usage:
#    iterativeStability(ExprsData, SampleGroups=NULL,
#     outputFilePrefix='IterativeStab_output', Niter=100)
#
#  Arguments:
#    ExprsData - data frame with expression estimates 
#      in log scale columns should represent samples, 
#      rows individual features e.g. genes, miRNAs.
#      Sample and feature labels should be defined in 
#      the col and row names respectively.
#    SampleGroups - optional vector that defines the 
#      sample group labels used in NormFinder
#    outputFilePrefix - prefix for the output file names
#    Niter - number of algorithm iterations which 
#      determines the maximum set of stable features
#
#  Output:
#    Excel file with the stability estimates for each 
#      tested feature or averaged set of features
#    
#
#########################################################

iterativeStability = function(ExprsData,SampleGroups=NULL,outputFilePrefix='IterativeStab_output',Niter=100) {
  
  require('plyr')
  require('ggplot2')
  require('openxlsx')  
  source('https://moma.dk/files/r.NormOldStab5.txt')
  
  TEMP_FILE='iterativeStabilityTempFilep.txt'
  
  #create data
  pData = ExprsData
  Nrows = dim(pData)[1]
  
  #create vector of available feature IDs
  mirIDXnr = data.frame(ID=1:Nrows)
  rownames(mirIDXnr) = rownames(pData)
  
  #create output and index tables
  ResTable = data.frame()
  TopMirS = NULL
  TopMirIDXS = NULL
  
  #limit the maximum number of iterations it it exceeds the total number of features
  if (Niter>Nrows) {Niter=Nrows}
  
  #main loop for adding new features 
  for (i in 1:Niter) {
    
    print(paste0('Processing iteration nr: ',i))
    ptm <- proc.time()
    
    #conduct the NormFinder analysis
    if (!is.null(SampleGroups)) {
      write.table(rbind(pData,SampleGroups),TEMP_FILE,sep="\t")
      NormfinderRes = Normfinder(TEMP_FILE,Groups=T,ctVal=T,pStabLim=0)
    } else {
      NormfinderRes = Normfinder(TEMP_FILE,Groups=F,ctVal=T,pStabLim=0)
    }

    #save detailed results of current iteration
    tResTable = NormfinderRes$Ordered
    tResTable$iter = i
    if (i>1) {
      rownames(tResTable) = paste(rownames(tResTable),paste(TopMirS,collapse=', '),sep=', ')  
    }
    ResTable = rbind(ResTable,tResTable)  
    
    #select the best feature
    TopMir = rownames(NormfinderRes$Ordered[1,])
    TopMirIDX = mirIDXnr[TopMir,1]
    
    #add the feature to the selected group
    TopMirS = c(TopMirS,TopMir)
    TopMirIDXS = c(TopMirIDXS,TopMirIDX)
    
    #create new data set by averaging the previously selected features with all that remain
    subMirIDXnr = mirIDXnr[!rownames(mirIDXnr) %in% TopMirS,]
    CombT=cbind(ID=subMirIDXnr,t(matrix(rep(TopMirIDXS,length(subMirIDXnr)),ncol =length(subMirIDXnr))))
    agrData = t(apply(CombT, 1, function(x) colMeans(ExprsData[unlist(x),])))
    rownames(agrData) = rownames(ExprsData)[subMirIDXnr]
    
    #move the new dataset to the next iteration
    pData = agrData
    elaptime = proc.time() - ptm
    print(paste0('   ...finished in: ',round(elaptime[3]/60,2),'min'))
  }
  
  #process the output table
  ResTable = cbind(ID = rownames(ResTable), ResTable)
  ResTable = ResTable[order(ResTable$Stability),]
  
  #creare the summary table for each iteration
  SummaryTable <- ddply(ResTable, 'iter', 
                   .fun = function(xx, col) {
                     c(N    = length(xx[[col]]),
                       mean = mean   (xx[[col]], na.rm=T),
                       med  = median (xx[[col]], na.rm=T),
                       min   = min     (xx[[col]], na.rm=T),
                       max   = max     (xx[[col]], na.rm=T)
                     )
                   },'Stability')

  
  #save all results to excel
  wb <- createWorkbook()
  addWorksheet(wb = wb, sheetName = "Iterations", gridLines = T);   writeDataTable(wb = wb, sheet = 1, x = ResTable)
  addWorksheet(wb = wb, sheetName = "Summary", gridLines = T);   writeDataTable(wb = wb, sheet = 2, x = SummaryTable)
  saveWorkbook(wb, paste0(outputFilePrefix,'.xlsx'), overwrite = TRUE)
  
  #create and save the plot
  p=ggplot(SummaryTable) + geom_line(aes(x=iter,y=min,colour='blue'))+
                       geom_line(aes(x=iter,y=mean,colour='darkgreen')) + 
                       geom_point(aes(x=iter,y=min,colour='blue')) + 
                       geom_point(aes(x=iter,y=mean,colour='darkgreen'))+
                       theme_bw() +ylab('NormFinder stability')  +xlab('number of features') +scale_x_continuous(breaks=1:Niter)+
                       scale_colour_manual(name = '', values =c('blue'='blue','darkgreen'='darkgreen'), labels = c('minimum','mean')) + 
                       theme(legend.position="bottom")
  ggsave(p,file=paste0(outputFilePrefix,'.png'), scale=2, width=2.5, height=2.5, dpi=300) 
  
  return(ResTable)
}